############################################################
#                       CGI-LIB.PL
#
# Perl Routinen zur Manipulierung von CGI-Eingaben
# cgi-lib@pobox.com
# $Id: cgi-lib.pl,v 2.17 1998/05/14 22:39:23 brenner Exp $
#
# Copyright (c) 1993-1998 Steven E. Brenner  
# Unpublished work.
# Permission granted to use and modify this library so long as the
# copyright above is maintained, modifications are documented, and
# credit is given for any use of the library.
#
# Diese Bibliothek darf solange benutzt und modifiziert werden,
# wie obiges Copryright aufgefuehrt wird und Aenderungen dokumentiert
# werden und die Benutzung der Bibliothek entsprechnd gewuerdigt
# wird.
#
# Dank gebuehrt einer Menge von Leuten, die Bugs berichtet und
# Vorschlaege gemacht haben.
#
# Deutsche Uebersetzung: Gerold Bigorajski
#     http://www.gb-direkt.de
#     email: support@gb-direkt.de
# Die Routinen RealName und RealPath wurden vom Uebersetzer 
# hinzugefuegt. Erlaeuterung siehe am Ende dieser Bibliothek.
#
#
# Weitere Informationen unter:
#     http://cgi-lib.stanford.edu/cgi-lib/

$cgi_lib'version = sprintf("%d.%02d", q$Revision: 2.17 $ =~ /(\d+)\.(\d+)/);


############################################################
# Parameter, die das cgi-lib-Verhalten beeinflussen.
# Benutzer-konfiguriebare Parameter bzgl. des Datei-Uploads.
############################################################
$cgi_lib'maxdata    = 131072;		# maximale Anzahl an Bytes, die mittels POST akzeptiert werden - 2^17.
$cgi_lib'writefiles = "/vserv/f/flenb01/htdocs/images";		# Verzeichnis, in das Dateien geschrieben werden sollen - oder 0,
						# falls keine Dateien geschrieben werden sollen.

$cgi_lib'filepre    = "";		# Praefix der Dateinamen im obigen Verzeichnis.

# Die folgenden Parameter sollten nicht veraendert werden, es sei denn, ein spezielle Grund dafuer liegt vor.
$cgi_lib'bufsize  =  8192;		# Standard-Puffergroesse beim Lesen von multipart.
$cgi_lib'maxbound =   100;		# Maximale Boundary-Laenge.
$cgi_lib'headerout =    0;		# Anzeige, ob der Header gedruckt worden ist.


############################################################
# Subroutine ReadParse
############################################################
# Liest mittels GET und POST uebermittelte Daten, konvertiert sie zu einfachem Text und erzeugt Paare
# aus Schluessel / Wert in %in, wobei "\0" verwendet wird, um mehrfache Selektionen zu trennen.
#
# Gibt > 0 zurueck, wenn es eine Eingabe gab und 0 im Falle keiner Eingabe. undef zeigt irgendeinen
# Fehler an.
#
# Da nun cgi-Scripte am gleichen Ort wie andere Dateien auch gespeichert werden koennen,
# ist es nuetzlich, das Formular und das Script an einer Stelle zu vereinen. Wenn keine
# Parameter angegeben sind, (ReadParse gibt FALSE zurueck) koennte die Ausgabe ein
# Formular sein.
#
# Wenn ein Verweis auf einen Hash angegeben wird, werden die Daten in diesem Hash gespeichert,
# aber auf die Daten in $in und @in kann dann nicht mehr zugegriffen werden.
# Wenn eine globale Variable (z.b. *cgi_input) der erste Paramter fuer ReadParse ist, werden
# Variablen dort statt in $in, @in und %in gespeichert.
# Der zweite, dritte und vierte Parameter fuellen Assoziative Arrays analog zu
# %in mit Daten bzgl. des Upload von Dateien. 
#
# Wenn keine Methode angegeben ist, verarbeitet das Script beide Befehlszeilenargumente
# des Formulars: Name=Wert und jeglicher Text, der in $ENV{'QUERY_STRING'} enthalten ist.
# Dies dient dazu, beim Debugging zu helfen und mag in zukuenftigen Versionen geaendert
# werden.

sub ReadParse
{
	local (*in) = shift if @_;    # CGI input
	local (*incfn,                # lokaler Dateiname bei Upload (muss nicht uebergeben werden)
		 *inct,                 # Content-type bei Upload (muss nicht uebergeben werden)
		 *insfn) = @_;          # Dateiname auf dem  Server (fuer gespoolte Dateien) (muss nicht uebergeben werden)
	local ($len, $type, $meth, $errflag, $cmdflag, $perlwarn, $got, $name);
	
	# Ausschalten von Warnmeldungen, da dieser Code freizuegig von lokalen und Umgebungsvariablen
	# Gebrauch macht, die zunaechst undefiniert sind (d.h.  explizite Initialisierung).
	$perlwarn = $^W;
	$^W = 0;

	binmode(STDIN);   # Dies wird fuer DOS-basierte Systeme gebraucht
	binmode(STDOUT);  # und sollte keinen weiteren Einfluss haben
	binmode(STDERR);
	
	# Einlesen einiger nuetzlicher Umgebungsvariablen:
	$type = $ENV{'CONTENT_TYPE'};
	$len  = $ENV{'CONTENT_LENGTH'};
	$meth = $ENV{'REQUEST_METHOD'};
  
	if ($len > $cgi_lib'maxdata)
	{ #'
		&CgiDie("cgi-lib.pl: Es wurden zu viele Daten angefordert: $len Bytes\n");
	}
  
	if (!defined $meth || $meth eq '' || $meth eq 'GET' || $meth eq 'HEAD' ||
		$type eq 'application/x-www-form-urlencoded')
	{
		local ($key, $val, $i);
	
		# Text einlesen
		if (!defined $meth || $meth eq '')
		{
			$in = $ENV{'QUERY_STRING'};
			$cmdflag = 1;  # ebenfalls Benutzung der Befehlszeilenparameter
		}
		elsif($meth eq 'GET' || $meth eq 'HEAD')
		{
			$in = $ENV{'QUERY_STRING'};
		}
		elsif ($meth eq 'POST')
		{
			if (($got = read(STDIN, $in, $len) != $len))
			{
				$errflag="Zu wenig Daten: erforderlich $len, erhalten $got\n";
			}
		}
		else
		{
			&CgiDie("cgi-lib.pl: Unbekannte Abfrage-Methode: $meth\n");
		}

		@in = split(/[&;]/,$in); 
		push(@in, @ARGV) if $cmdflag; # Befehlszeilen-Parameter einfuegen

		foreach $i (0 .. $#in)
		{
			# Pluszeichen in Leerzeichen konvertieren.
      		$in[$i] =~ s/\+/ /g;

			# In Schluessel und Wert aufspalten.  
			($key, $val) = split(/=/,$in[$i],2); # splits on the first =.

			# Konvertieren der Hexwerte %XX in alphanumerische.
			$key =~ s/%([A-Fa-f0-9]{2})/pack("c",hex($1))/ge;
			$val =~ s/%([A-Fa-f0-9]{2})/pack("c",hex($1))/ge;

			# Schluessel und Wert miteinander verknuepfen.
			$in{$key} .= "\0" if (defined($in{$key})); # \0 ist das multiple Trennzeichen
			$in{$key} .= $val;
		}

	} # Ende der Abfrage if (!defined $meth || $meth eq '' || $meth eq 'GET' || $type eq 'application/x-www-form-urlencoded')
	elsif ($ENV{'CONTENT_TYPE'} =~ m#^multipart/form-data#)
	{
	# Aus Effinziengruenden wird Multipart-Code nur im Bedarfsfalle kompliliert
	$errflag = !(eval <<'END_MULTIPART');

	local ($buf, $boundary, $head, @heads, $cd, $ct, $fname, $ctype, $blen);
	local ($bpos, $lpos, $left, $amt, $fn, $ser);
	local ($bufsize, $maxbound, $writefiles) = ($cgi_lib'bufsize, $cgi_lib'maxbound, $cgi_lib'writefiles);

	# Die folgende Zeile dient einfach dazu, diverse Warnmeldungen zu eliminieren:
	$buf = ''; 

	($boundary) = $type =~ /boundary="([^"]+)"/; #";   # Boundary finden
	($boundary) = $type =~ /boundary=(\S+)/ unless $boundary;
	&CgiDie ("Boundary wird nicht untersttzt: Wahrscheinlich ein Fehler Ihres Servers") unless $boundary;
	$boundary =  "--" . $boundary;
	$blen = length ($boundary);

	if ($ENV{'REQUEST_METHOD'} ne 'POST')
	{
		&CgiDie("Ungltige Request-Methode fr multipart/form-data: $meth\n");
	}

	if ($writefiles)
	{
		local($me);
		stat ($writefiles);
		$writefiles = "/tmp" unless  -d _ && -w _;
		# ($me) = $0 =~ m#([^/]*)$#;
		$writefiles .= "/$cgi_lib'filepre"; 
	}
	# Einlesen der Daten und Aufsplitten in ihre Bestandteile:
	# Header kommen in @in und Daten in %in
	# Allgemeiner Algorithmus:
	#   Es gibt zwei Trennzeichen. Der Rand und '\r\n\r\n' zwischen
	#   Header und Body. Abwechselnde Suche nach ihnen.
	#   Bereitstellen eines Puffers der Groesse (bufsize+maxbound); letzteres
	#   deswegen, damit Trennzeichen nicht durch Aufteilung zwischen zwei
	#   Puffern verloren gehen.
	#   Suche nach einem Trennzeichen im augenblicklichen Stapel. Wenn keins
	#   gefunden wird, Sichern des gesamten Puffers, den maxbound-Extrapuffer
	#   an den Anfang des Puffers verschieben und die naechsten bufsize Bytes
	#   einlesen.
	#   Wenn ein Trennzeichen gefunden wurde, alles bis zum Trennzeichen sichern.
	#   Leeren des Puffers bis zum Ende des Trennzeichens.
	#   Wiederauffuellen des Puffers bis bufsize+maxbound.
	# Beachten Sie die etwas verquere Organisation: Der Code vor BODY: gehoert
	# wirklich zu dem Code, der auf HEAD: folgt, wird aber zuerst genommen, um
	# den Puffer 'vorab zu fuellen'. BODY: steht vor HEAD:, weil zuerst jeglicher
	# 'Vorspann' geloescht werden muss, der einem Body ohne vorausgehenden Head
	# entspraeche.

	$left = $len;
	PART: # Auffinden jedes Teils der Multipart-Daten waehrend des Einlesens
	while (1)
	{
		die $@ if $errflag;

		$amt = ($left > $bufsize+$maxbound-length($buf) ? $bufsize+$maxbound-length($buf): $left);
		$errflag = (($got = read(STDIN, $buf, $amt, length($buf))) != $amt);
		die "Zu wenig Daten: erforderlich $amt, erhalten $got\n" if $errflag;
		$left -= $amt;

		$in{$name} .= "\0" if defined $in{$name}; 
		$in{$name} .= $fn if $fn;

		$name=~/([-\w]+)/;  # Damit kann $insfn{$name} unberuehrt bleiben.
		if (defined $1)
		{
			$insfn{$1} .= "\0" if defined $insfn{$1}; 
			$insfn{$1} .= $fn if $fn;
		}
 
		BODY: 
		while (($bpos = index($buf, $boundary)) == -1)
		{
			if ($left == 0 && $buf eq '')
			{
				foreach $value (values %insfn)
				{
					unlink(split("\0",$value));
				}
				&CgiDie("cgi-lib.pl: Bei der Suche nach der Multipart-Boundary wurde das Ende der Eingabe erreicht" .
					"Fehlerhaftes Format der CGI-Eingabe.\n");
			}
			die $@ if $errflag;
			if ($name)
			{  # wenn es kein $name, dann ist es der Vorspann - ergo loeschen.
				if ($fn)
				{
					print FILE substr($buf, 0, $bufsize);
				}
				else
				{
					$in{$name} .= substr($buf, 0, $bufsize);
				}
			}
			$buf = substr($buf, $bufsize);
			$amt = ($left > $bufsize ? $bufsize : $left); #$maxbound==length($buf);
			$errflag = (($got = read(STDIN, $buf, $amt, length($buf))) != $amt);
			die "Zu wenig Daten: erforderlich $amt, erhalten $got\n" if $errflag;
			$left -= $amt;
		} # Ende der Schleife while (($bpos = index($buf, $boundary)) == -1)

		if (defined $name)
		{  # Wenn es kein $name gibt, ist es der Vorspann - ergo loeschen.
			if ($fn)
			{
				print FILE substr($buf, 0, $bpos-2);
			}
			else
			{
				$in {$name} .= substr($buf, 0, $bpos-2); } # Loeschen des letzten \r\n
			}
			close (FILE);
			last PART if substr($buf, $bpos + $blen, 2) eq "--";
			substr($buf, 0, $bpos+$blen+2) = '';
			$amt = ($left > $bufsize+$maxbound-length($buf) ? $bufsize+$maxbound-length($buf) : $left);
			$errflag = (($got = read(STDIN, $buf, $amt, length($buf))) != $amt);
			die "Zu wenig Daten: erforderlich $amt, erhalten $got\n" if $errflag;
			$left -= $amt;

			undef $head;
			undef $fn;

			HEAD:
			while (($lpos = index($buf, "\r\n\r\n")) == -1)
			{ 
				if ($left == 0  && $buf eq '')
				{
					foreach $value (values %insfn)
					{
						unlink(split("\0",$value));
					}
					&CgiDie("cgi-lib: Bei der Suche nach dem Ende des Headers wurde das Ende der Eingabe erreicht. " .
						"Falsches Format der CGI-Eingabe.\n$buf");
				}
				die $@ if $errflag;
				$head .= substr($buf, 0, $bufsize);
				$buf = substr($buf, $bufsize);
				$amt = ($left > $bufsize ? $bufsize : $left); #$maxbound==length($buf);
				$errflag = (($got = read(STDIN, $buf, $amt, length($buf))) != $amt);
				die "Zu wenig Daten: erforderlich $amt, erhalten $got\n" if $errflag;
				$left -= $amt;
			}
			$head .= substr($buf, 0, $lpos+2);
			push (@in, $head);
			@heads = split("\r\n", $head);
			($cd) = grep (/^\s*Content-Disposition:/i, @heads);
			($ct) = grep (/^\s*Content-Type:/i, @heads);

			($name) = $cd =~ /\bname="([^"]+)"/i; #"; 
			($name) = $cd =~ /\bname=([^\s:;]+)/i unless defined $name;  

			($fname) = $cd =~ /\bfilename="([^"]*)"/i; #"; # Der Dateiname kann ein leerer String sein
			($fname) = $cd =~ /\bfilename=([^\s:;]+)/i unless defined $fname;
			$incfn{$name} .= (defined $in{$name} ? "\0" : "") . (defined $fname ? $fname : "");

			($ctype) = $ct =~ /^\s*Content-type:\s*"([^"]+)"/i;  #";
			($ctype) = $ct =~ /^\s*Content-Type:\s*([^\s:;]+)/i unless defined $ctype;
			$inct{$name} .= (defined $in{$name} ? "\0" : "") . $ctype;

			if ($writefiles && defined $fname)
			{
				$ser++;
				$fn = $writefiles . ".$$.$ser";
				open (FILE, ">$fn") || &CgiDie("Konnte $fn nicht ffnen\n");
				binmode (FILE);  # Akkurates Schreiben der Dateien
			}
			substr($buf, 0, $lpos+4) = '';
			undef $fname;
			undef $ctype;
		}

1;
END_MULTIPART
		if ($errflag)
		{
			local ($errmsg, $value);
			$errmsg = $@ || $errflag;
			foreach $value (values %insfn)
			{
				unlink(split("\0",$value));
			}
			&CgiDie($errmsg);
		}
		else
		{
			# Alles ist in Ordnung.
		}
	}
	else
	{
		&CgiDie("cgi-lib.pl: Unbekannter Content-type: $ENV{'CONTENT_TYPE'}\n");
	}

	# Zur Vermeidung von Warnmeldungen ('used only once')
	$insfn = $insfn;
	$incfn = $incfn;
	$inct  = $inct;

	$^W = $perlwarn;

	return ($errflag ? undef :  scalar(@in)); 
}


############################################################
# Subroutine PrintHeader
############################################################
# 
# gibt die magische zeile zurueck, die dem Browser mitteilt, dass ein
# HTML-Dokument folgt.

sub PrintHeader
{
	return "Content-type: text/html\n\n";
}


############################################################
# Subroutine HtmlTop
############################################################
# 
# Gibt den <head> eines Dokuments und den Anfang des body zurueck -
# mit dem titel und einem body <h2>-Header, wie es vom Parameter
# spezifiziert wird.

sub HtmlTop
{
	local ($title, $bodytag, $fonttag) = @_;

	return <<END_OF_TEXT;
<html>
<head>
<title>$title</title>
</head>
<body $bodytag>
<table align="center" width="100\%">
	<tr>
		<td align="left"><h2>$fonttag$title</font></h2></td>
		<td align="right" valign="middle"><p>$fonttag<a href="javascript:history.back()">Zurck</a></font></td>
	</tr>
</table>
END_OF_TEXT
}


############################################################
# Subroutine HtmlBot
############################################################
# 
# Gibt den abschliessenden </body>, </html> -Code fuer jede HTML-Seite zurueck.

sub HtmlBot
{
	return "</body>\n</html>\n";
}


############################################################
# Subroutine SplitParam
############################################################
# 
# Splittet einen Parameter mit vielfachen Werten in eine Liste
# der ihn ausmachenden Parameter.

sub SplitParam
{
	local ($param) = @_;
	local (@params) = split ("\0", $param);
	return (wantarray ? @params : $params[0]);
}


############################################################
# Subroutine MethGet
############################################################
# 
# Liefert true, falls der cgi-Aufruf mittels GET erfolgte, sonst false.

sub MethGet
{
	return (defined $ENV{'REQUEST_METHOD'} && $ENV{'REQUEST_METHOD'} eq "GET");
}


############################################################
# Subroutine MethPost
############################################################
# 
# Liefert true, falls der cgi-Aufruf mittels POST erfolgte, sonst false.

sub MethPost
{
	return (defined $ENV{'REQUEST_METHOD'} && $ENV{'REQUEST_METHOD'} eq "POST");
}


############################################################
# Subroutine MyBaseUrl
############################################################
# 
# Gibt die Basis-URL des Scripts zurueck (ohne Extra-Pfad
# oder Abfrage).

sub MyBaseUrl
{
	local ($ret, $perlwarn);
	$perlwarn = $^W;
	$^W = 0;
	$ret = 'http://' . $ENV{'SERVER_NAME'} . ($ENV{'SERVER_PORT'} != 80 ? ":$ENV{'SERVER_PORT'}" : '') . $ENV{'SCRIPT_NAME'};
	$^W = $perlwarn;
	return $ret;
}


############################################################
# Subroutine MyFullUrl
############################################################
# 
# Gibt die vollstaendige URL des Scripts zurueck (mit Extra-Pfad
# und Abfrage).

sub MyFullUrl
{
	local ($ret, $perlwarn);
	$perlwarn = $^W; $^W = 0;
	$ret = 'http://' . $ENV{'SERVER_NAME'} . ($ENV{'SERVER_PORT'} != 80 ? ":$ENV{'SERVER_PORT'}" : '') . 
         $ENV{'SCRIPT_NAME'} . $ENV{'PATH_INFO'} . (length ($ENV{'QUERY_STRING'}) ? "?$ENV{'QUERY_STRING'}" : '');
	$^W = $perlwarn;
	return $ret;
}


############################################################
# Subroutine MyUrl
############################################################
# 
# Verweist den Browser zurueck auf das Script. Dies ist veraltet
# und wird in spaeteren Versionen entfernt werden.

sub MyURL
{
	return &MyBaseUrl;
}


############################################################
# Subroutine CgiError
############################################################
# 
# Gibt eine Fehlermeldung aus, die angemessene Header usw. enthaelt.
# Parameter:
#  Ohne Paramter Ausgabe einer allgemeinen Fehlermeldung
#  Ansonsten wird der erste Parameter der Titel und der Rest wird zu
#  verschiedenen Absaetzen des Bodies.

sub CgiError
{
	local (@msg) = @_;
	local ($i,$name);

	if (!@msg)
	{
		$name = &MyFullUrl;
		@msg = ("Fehler: Das Script $name produzierte einen fatalen Fehler\n");
	}

	if (!$cgi_lib'headerout)
	{ #')
		print &PrintHeader;	
		print "<html>\n<head>\n<title>$msg[0]</title>\n</head>\n<body>\n";
	}
	print "<h2>$msg[0]</h2>\n";
	foreach $i (1 .. $#msg)
	{
		print "<p>$msg[$i]</p>\n";
	}

	$cgi_lib'headerout++;
}


############################################################
# Subroutine CgiDie
############################################################
# 
# Identisch zu CgiError, beendet aber auch zugleich mit der Fehlermeldung
# das Progamm.

sub CgiDie
{
	local (@msg) = @_;
	&CgiError (@msg);
	die @msg;
}


############################################################
# Subroutine PrintVariables
############################################################
# 
# Huebsch formatierte Variable. Drei Aufruf-Optionen:
# Ein nicht-leeres assoziatives Array - druckt die Bestandteile dieses Arrays.
# Ein globaler type - druckt die Bestandteile des assoziativen Arrays.
# Nichts - Standardwerte, die in %in zu verwenden.
# Typische Verwendung: &PrintVariables()

sub PrintVariables
{
	local (*in) = @_ if @_ == 1;
	local (%in) = @_ if @_ > 1;
	local ($out, $key, $output);

	$output =  "\n<dl compact>\n";
	foreach $key (sort keys(%in))
	{
		foreach (split("\0", $in{$key}))
		{
			($out = $_) =~ s/\n/<br>\n/g;
			$output .=  "<dt><b>$key</b>\n <dd>:<i>$out</i>:<br>\n";
		}
	}
	$output .=  "</dl>\n";

	return $output;
}


############################################################
# Subroutine PrintEnv
############################################################
# 
# Formatiert alle Umgebungsvariable anstaendig und gibt
# einen HTML-String zurueck.

sub PrintEnv
{
	&PrintVariables(*ENV);
}


############################################################
# Subroutine RealName
############################################################
# 
# Ermittelt den tatsaechlichen Dateinamen aus der 
# vollstaendigen Upload-Pfadangabe.

sub RealName
{
	local($param) = $_[0];
	local($max)   = 0;
	local($output);

	# Bestimmen des letzten Auftretens von "\" (Windows), "/" (UNIX), ":" (Apple)
	# und Setzen auf das Maximum der drei.
	$max = rindex($param, "\\");
	$max = $max > rindex($param, "\/") ? $max : rindex($param, "\/");
	$max = $max > rindex($param, "\:") ? $max : rindex($param, "\:");

	# Der wirkliche Dateiname ist das, was nach dem letzten Auftreten
	# eines dieser drei Zeichen folgt
	$output = substr($param, $max + 1);

	return $output;
}


############################################################
# Subroutine RealPath
############################################################
# 
# Ermittelt den tatsaechlichen Pfad (ohne Dateiname)
# aus der vollstaendigen Upload-Pfadangabe.
# Das abschliessende Pfad-Trennzeichen wird unterdrueckt

sub RealPath
{
	local($param) = $_[0];
	local($max)   = 0;
	local($output);

	# Bestimmen des letzten Auftretens von "\" (Windows), "/" (UNIX), ":" (Apple)
	# und Setzen auf das Maximum der drei.
	$max = rindex($param, "\\");
	$max = $max > rindex($param, "\/") ? $max : rindex($param, "\/");
	$max = $max > rindex($param, "\:") ? $max : rindex($param, "\:");

	# Der wirkliche Dateiname ist das, was nach dem letzten Auftreten
	# eines dieser drei Zeichen folgt
	$output = substr($param, 0, $max);

	return $output;
}


# Die folgenden Zeilen existieren nur, um Warnmeldungen zu vemeiden
$cgi_lib'writefiles =  $cgi_lib'writefiles;
$cgi_lib'bufsize    =  $cgi_lib'bufsize ;
$cgi_lib'maxbound   =  $cgi_lib'maxbound;
$cgi_lib'version    =  $cgi_lib'version;
$cgi_lib'filepre    =  $cgi_lib'filepre;

# Als Bibliothek wird immer true zurueckgegeben

1;

